% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{print.cbtf}
\alias{print.cbtf}
\title{Print the results from a fuzz run}
\usage{
\method{print}{cbtf}(x, show = c("fail", "warn"), group = "input", ...)
}
\arguments{
\item{x}{An object of class \code{cbtf}.}

\item{show}{A character vector representing the subset of results to be
printed, any of "fail", "warn", "skip", "ok", "all" and "none".}

\item{group}{Either \code{"input"} to show results grouped by test input
(default), or \code{"function"} to show them grouped by function name.}

\item{...}{Further arguments passed to or from other methods.
These are currently ignored.}
}
\value{
No return value, called for side effects.
}
\description{
This formats the results from a fuzz run with colours and prints them to
the terminal. By default, only tests labelled as \code{"FAIL"} or \code{"WARN"} are
displayed, but this can be controlled via the \code{show} argument.
}
\details{
The use of unicode icons in the output messages can be disabled by setting
\code{options(cli.unicode = FALSE)}.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
print(res)
print(res, show = "all")
print(res, show = "none")
print(res, group = "function")

}
\seealso{
\link{summary.cbtf}
}
