% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{length.cbtf}
\alias{length.cbtf}
\title{Compute the number of tests performed}
\usage{
\method{length}{cbtf}(x)
}
\arguments{
\item{x}{An object of class \code{cbtf}.}
}
\value{
An integer corresponding to the number of tests performed in a run.
}
\description{
This returns the total number of tests as the product of the number of
inputs used and the number of functions tested.
}
\examples{
res <- fuzz(funs = c("list", "matrix", "mean"),
            what = test_inputs(c("numeric", "raw")))
length(res)

}
